from pathlib import Path
from typing import List, Type, Optional

from autograder.util import AutograderError, import_from_path
from .abstract_testcase import TestCase
from .stdout_testcase import StdoutOnlyTestCase


class TestCasePicker:
    testcase_types: List[Type[TestCase]]

    def __init__(
        self,
        testcase_types_dir: Path,
        possible_source_file_stems: List[str],
        stdout_only_grading_enabled: bool = False,
    ):
        self.possible_source_file_stems = possible_source_file_stems
        self.testcase_types = self.discover_testcase_types(testcase_types_dir)
        if stdout_only_grading_enabled:
            self.testcase_types.insert(0, StdoutOnlyTestCase)
        if not self.testcase_types:
            raise AutograderError("No acceptable testcase types were detected.\n")

    @classmethod
    def discover_testcase_types(cls, testcase_types_dir: Path) -> List[Type[TestCase]]:
        testcase_types = []
        for testcase_type_dir in testcase_types_dir.iterdir():
            for path in testcase_type_dir.iterdir():
                if path.is_file() and path.suffix == ".py":
                    module = import_from_path(f"testcase:{path.stem}{testcase_type_dir.name}", path)
                    testcase_type = getattr(module, "TestCase", None)
                    if testcase_type is None:
                        print(f"Testcase type '{path}' does not define a 'TestCase' class, skipping.")
                        continue

                    if cls._is_installed(testcase_type_dir.name, testcase_type):
                        testcase_types.append(testcase_type)
        return testcase_types

    @staticmethod
    def _is_installed(language_name: str, testcase: Type[TestCase]) -> bool:
        """Useful for logging"""
        if testcase.is_installed():
            return True
        else:
            print(f"Utilities for running {language_name} are not installed. Disabling it.")
            return False

    def pick(self, file: Path) -> Optional[Type[TestCase]]:
        for testcase_type in self.testcase_types:
            if testcase_type.is_a_type_of(file, self.possible_source_file_stems):
                return testcase_type



from pathlib import Path
from typing import List, Optional, Type, Dict, Tuple

from .abstract_testcase import TestCase


def find_appropriate_source_file_stem(file: Path, possible_source_file_stems: List[str]) -> Optional[str]:
    for s in possible_source_file_stems:
        if s in file.stem:
            return s


class Submission:
    # I wanted to use a dataclass but they do not work well with slots until 3.10
    __slots__ = "path", "type", "dir", "grades"

    file: Path
    type: Type[TestCase]
    dir: Path

    grades: Dict[str, Tuple[float, str]]

    def __init__(self, file: Path, testcase_type: Type[TestCase], temp_dir: Path):
        self.path = file
        self.type = testcase_type
        self.dir = temp_dir / file.name
        self.dir.mkdir()
        self.grades = {}

    def add_grade(self, test_name: str, testcase_score: float, message: str):
        self.grades[test_name] = (testcase_score, message)



from pathlib import Path

FILE_WITH_TOKEN_INFO = Path(__file__).parent / "token.txt"


class Token:
    base_url: str
    token: str
    file: Path

    def __init__(self, base_url, token, file_with_token_info: Path):
        self.base_url = base_url.strip()
        self.token = token.strip()
        self.file = file_with_token_info

    @classmethod
    def create(cls, base_url: str, file_with_token_info: Path) -> "Token":
        token = cls(base_url, cls._generate_new_token(base_url), file_with_token_info)
        token.save()
        return token

    @classmethod
    def from_file(cls, file_with_token_info: Path) -> "Token":
        with open(file_with_token_info) as f:
            base_url, token = f
        return cls(base_url, token, file_with_token_info)

    def save(self):
        with open(self.file, "w") as f:
            f.write(f"{self.base_url}\n{self.token}")

    def delete(self):
        self.file.unlink()

    @staticmethod
    def _generate_new_token(base_url: str) -> str:
        # TODO
        raise NotImplementedError



import enum
import shutil
from abc import ABC, abstractmethod, ABCMeta
from inspect import getsourcefile
from io import StringIO
from pathlib import Path
from typing import Dict
from typing import List, Tuple

import sh

from .exit_codes import ExitCodeEventType, USED_EXIT_CODES, SYSTEM_RESERVED_EXIT_CODES
from .shell import get_stderr, ShCommand
from .test_helper_formatter import get_formatted_test_helper
from .testcase_io import TestCaseIO
from .testcase_result_validator import generate_validating_string, validate_output


# TODO: Why is this not in config?
class ArgList(enum.Enum):
    SUBMISSION_PRECOMPILATION = "SUBMISSION_PRECOMPILATION_ARGS"
    TESTCASE_PRECOMPILATION = "TESTCASE_PRECOMPILATION_ARGS"
    TESTCASE_COMPILATION = "TESTCASE_COMPILATION_ARGS"


class SourceDirSaver(ABCMeta, type):
    """Useful in getting the resources associated with each testcase type"""

    type_source_file: Path

    def __new__(mcs, name, bases, dct):
        cls = super().__new__(mcs, name, bases, dct)
        # We needed a way to get a source file based solely on __class__ to access its sibling directories
        source_file = getsourcefile(cls)
        if source_file is None:
            raise FileNotFoundError(f"Source file for class {cls} has not been found.")
        cls.type_source_file = Path(source_file)
        return cls


class TestCase(ABC, metaclass=SourceDirSaver):
    source_suffix = ".source_suffix"  # dummy value
    executable_suffix = ".executable_suffix"  # dummy value

    type_source_file: Path
    test_helpers_dir: Path
    path: Path
    weight: float
    max_score: int
    io: TestCaseIO
    validating_string: str

    @property
    @abstractmethod
    def helper_module(self) -> str:
        pass

    @classmethod
    @abstractmethod
    def is_installed(cls) -> bool:
        """Returns True if software necessary to run the testcase is installed on the system"""

    @abstractmethod
    def compile_testcase(self, precompiled_submission: Path) -> ShCommand:
        """Compiles student submission and testcase into a single executable
        (or simply returns the command to run the testcase if no further compilation is necessary)

        pwd = temp/student_dir
        """

    @classmethod
    def get_template_dir(cls):
        return cls.type_source_file.parent / "templates"

    def __init__(
        self,
        path: Path,
        timeout: float,
        argument_lists: Dict[ArgList, List[str]],
        weight: float,
        io: TestCaseIO,
    ):
        self.test_helpers_dir = self.type_source_file.parent / "helpers"
        self.path = path
        self.timeout = timeout
        self.argument_lists = argument_lists
        self.weight = weight
        self.max_score = int(weight * 100)

        # Only really works if test name is in snake_case
        self.name = path.stem.replace("_", " ").capitalize()

        self.io = io
        self.validating_string = generate_validating_string()

        self.prepend_test_helper()
        self.precompile_testcase()

    @classmethod
    def precompile_submission(
        cls,
        submission: Path,
        student_dir: Path,
        possible_source_file_stems: List[str],
        arglist: List[str],
    ) -> Path:
        """Copies student submission into student_dir and either precompiles
        it and returns the path to the precompiled submission or to the
        copied submission if no precompilation is necessary

        pwd = temp/student_dir
        """
        destination = (student_dir / possible_source_file_stems[0]).with_suffix(cls.source_suffix)
        shutil.copy(str(submission), str(destination))
        return destination

    def precompile_testcase(self):
        """Replaces the original testcase file with its compiled version,
        thus making reading its contents as plaintext harder.
        Useful in preventing cheating.

        pwd = AutograderPaths.current_dir (i.e. the directory with all submissions)
        """

    @classmethod
    def run_additional_testcase_operations_in_student_dir(cls, student_dir: Path):
        """Do nothing by default"""
        pass

    @classmethod
    def is_a_type_of(cls, file: Path, possible_source_file_stems: List[str]) -> bool:
        return file.suffix == cls.source_suffix

    def get_path_to_helper_module(self) -> Path:
        return self.test_helpers_dir / self.helper_module

    def run(self, precompiled_submission: Path) -> Tuple[float, str]:
        """Returns student score and message to be displayed"""
        result, message = self._weightless_run(precompiled_submission)

        self.delete_executable_files(precompiled_submission)
        return result * self.weight, message

    def make_executable_path(self, submission: Path) -> Path:
        """By combining test name and student name, it makes a unique path"""
        return submission.with_name(self.path.stem + submission.stem + self.executable_suffix)

    def prepend_test_helper(self):
        """Prepends all of the associated test_helper code to test code

        pwd = AutograderPaths.current_dir (i.e. the directory with all submissions)
        """
        with self.path.open() as f:
            content = f.read()
            final_content = self.get_formatted_test_helper() + "\n" + content
        with self.path.open("w") as f:
            f.write(final_content)

    def get_formatted_test_helper(self) -> str:
        return get_formatted_test_helper(self.get_path_to_helper_module())

    def delete_executable_files(self, precompiled_submission: Path):
        path = self.make_executable_path(precompiled_submission)
        if path.exists():
            path.unlink()

    def delete_source_file(self, source_path: Path):
        if source_path.exists():
            source_path.unlink()

    def _weightless_run(self, precompiled_submission: Path) -> Tuple[float, str]:
        """Returns student score (without applying testcase weight) and message to be displayed"""
        testcase_copy_in_student_dir = precompiled_submission.with_name(self.path.name)
        shutil.copy(str(self.path), str(testcase_copy_in_student_dir))

        try:
            test_executable = self.compile_testcase(precompiled_submission)
        except sh.ErrorReturnCode as e:
            return 0, get_stderr(e, "Failed to compile")
        self.delete_source_file(testcase_copy_in_student_dir)

        with StringIO() as runtime_output, self.io.input() as runtime_input:
            try:
                result = test_executable(
                    _in=runtime_input,
                    _out=runtime_output,
                    _timeout=self.timeout,
                    _ok_code=USED_EXIT_CODES,
                    _env={"VALIDATING_STRING": self.validating_string},
                )
                if result is None:
                    raise NotImplementedError()
                exit_code = result.exit_code
            except sh.TimeoutException:
                return 0, "Exceeded Time Limit"
            except sh.ErrorReturnCode as e:
                return (
                    0,
                    f"Crashed due to signal {e.exit_code}:\n{e.stderr.decode('UTF-8', 'replace')}\n",
                )
            raw_output = runtime_output.getvalue()
            output, score, output_is_valid = validate_output(raw_output, self.validating_string)
            if not output_is_valid:
                # This  means that either the student used built-in exit function himself
                # or some testcase helper is broken, or a testcase exits itself without
                # the use of helper functions.
                return (
                    0,
                    "None of the helper functions have been called.\n"
                    f"Instead, exit() has been called with exit_code {exit_code}.\n"
                    "It could indicate student cheating or testcase_utils being written incorrectly.",
                )
            elif exit_code == ExitCodeEventType.CHECK_STDOUT:
                if self.io.expected_output_equals(output):
                    return 100, f"{int(100 * self.weight)}/{self.max_score}"
                else:
                    return 0, f"0/{self.max_score} (Wrong output)"
            elif exit_code == ExitCodeEventType.RESULT:
                message = f"{round(score * self.weight, 2)}/{self.max_score}"
                if score == 0:
                    message += " (Wrong answer)"
                return score, message
            elif exit_code in SYSTEM_RESERVED_EXIT_CODES or exit_code < 0:
                # We should already handle this case in try, except block. Maybe we need more info in the error?
                raise NotImplementedError(f"System error with exit code {exit_code} has not been handled.")
            else:
                raise ValueError(f"Unknown system code {exit_code} has not been handled.")



import sys
from pathlib import Path
import argparse

from autograder.canvas.auth import Token, FILE_WITH_TOKEN_INFO
from autograder.canvas.submissions import upload_results


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    return _evaluate_args(_create_parser().parse_args(argv))


def _create_parser():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(title="Commands", dest="api_key_command")
    _create_upload_parser(subparsers)
    _create_api_key_parser(subparsers)
    return parser


def _create_upload_parser(subparsers: argparse._SubParsersAction):
    parser = subparsers.add_parser("upload", help="Upload assignments back to canvas")
    parser.add_argument(
        "submission_path",
        type=Path,
        nargs="?",
        default=Path.cwd(),
        help="Path to directory that contains student submissions",
    )
    parser.add_argument("course_id", type=int)
    parser.add_argument("assignment_id", type=int)
    parser.add_argument(
        "-g",
        "--grade",
        action="store_true",
        help="Also upload grades for each student submission.",
    )


def _create_api_key_parser(subparsers: argparse._SubParsersAction):
    parser = subparsers.add_parser("token", help="Create/Save/Get/Delete api_key")
    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        "-c",
        "--create",
        action="store",
        metavar="<base_url>",
        help="Create a new canvas api key and save it.",
    )
    group.add_argument("-g", "--get", action="store_true", help="Print the api key saved for the specified url.")
    group.add_argument("-d", "--delete", action="store_true", help="Delete the api key saved for the specified url.")
    group.add_argument(
        "-s",
        "--save",
        action="store",
        nargs=2,
        metavar=("<base_url>", "<token>"),
        help="Save the specified api-token globally.",
    )


def _evaluate_args(args):
    if args.api_key_command == "token":
        if args.create:
            return Token.create(args.create, FILE_WITH_TOKEN_INFO)
        elif args.save:
            Token(args.save[0], args.save[1], FILE_WITH_TOKEN_INFO).save()
        elif args.get:
            token = Token.from_file(FILE_WITH_TOKEN_INFO)
            print(f"Base url: {token.base_url}, Token: {token.token}")
            return token
        elif args.delete:
            Token.from_file(FILE_WITH_TOKEN_INFO).delete()
    elif args.api_key_command == "upload":
        upload_results(args.submission_path, args.course_id, args.assignment_id, args.grade)


if __name__ == "__main__":
    main()



import argparse
import sys
from pathlib import Path


def main(argv=None):
    """Returns the average score of the students"""
    if argv is None:
        argv = sys.argv[1:]

    parser = _create_parser()

    args = parser.parse_args(argv)
    if args.version:
        from autograder.__version__ import __version__

        print(__version__)
        exit(0)
    current_dir = (Path.cwd() / args.submission_path).resolve()
    return _evaluate_args(args, current_dir)


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--version", action="store_true", help="print the autograder version number and exit")
    subparsers = parser.add_subparsers(title="Commands", dest="command")
    _create_run_parser(subparsers)
    _create_stats_parser(subparsers)
    _create_guide_parser(subparsers)
    return parser


def _create_run_parser(subparsers):
    parser = subparsers.add_parser("run", help="Grade submissions in submission path or in current directory")
    parser.add_argument("--no_output", action="store_true", help="Do not output any code to the console")
    parser.add_argument(
        "-s",
        "--submissions",
        action="store",
        nargs="*",
        metavar="<name>",
        default=[],
        help="Only grade submissions with specified file names (without full path)",
    )
    _add_submission_path_argument(parser)


def _create_stats_parser(subparsers):
    parser = subparsers.add_parser("stats", help="Display statistics on student grades")
    parser.add_argument(
        "-p",
        "--print",
        type=float,
        nargs="?",
        default=None,
        const=100,
        metavar="min_score",
        help="Use after already graded to print assignments with score >= min_score",
    )
    _add_submission_path_argument(parser)


def _create_guide_parser(subparsers):
    parser = subparsers.add_parser("guide", help="Guide you through setting up a grading environment")
    _add_submission_path_argument(parser)


def _add_submission_path_argument(parser: argparse.ArgumentParser):
    parser.add_argument(
        "submission_path",
        type=Path,
        nargs="?",
        default=Path.cwd(),
        help="Path to directory that contains student submissions",
    )


def _evaluate_args(args, current_dir):
    if sys.platform.startswith("win32"):
        print(
            "Windows is not supported by autograder. If you do not have Linux,"
            "try using it through utilities like Windows Subsystem For Linux."
        )
        exit(1)
    elif sys.platform.startswith("darwin"):
        print("OSX is not officially supported. Proceed with caution.")
    from autograder.autograder import Grader, AutograderPaths  # That's some awful naming
    from autograder.util import AutograderError, print_results

    try:
        if args.command == "stats":
            if args.print:
                print_results(AutograderPaths(current_dir), args.print)
        elif args.command == "guide":
            from autograder import guide

            guide.main(AutograderPaths(current_dir))
        elif args.command == "run":
            return Grader(current_dir, no_output=args.no_output, submissions=args.submissions).run()
        else:
            raise NotImplementedError(f"Unknown command '{args.command}' supplied.")
    except AutograderError as e:
        print(e)
    return -1


if __name__ == "__main__":
    main()



from pathlib import Path
from canvasapi import Canvas
from .auth import Token, FILE_WITH_TOKEN_INFO

import shutil


def upload_results(results_dir: Path, course_id: int, assignment_id: int, grade: bool):
    # TODO
    print(f"Upload({results_dir}, {course_id}, {assignment_id}, {grade})")
    raise NotImplementedError
    shutil.make_archive(str(results_dir.with_suffix(".zip")), "zip", results_dir)
    token = Token.from_file(FILE_WITH_TOKEN_INFO)
    api = Canvas(token.base_url, token.token)
    submissions = api.get_course(course_id).get_assignment(assignment_id).get_submissions()
